SystemOrganization addCategory: #'Zinc-WebSocket-Core'!

Error subclass: #ZnWebSocketFailed
	instanceVariableNames: 'response'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-WebSocket-Core'!

!ZnWebSocketFailed commentStamp: '<historical>' prior: 0!
I am ZnWebSocketFailed, an exception thrown when a client WebSocket cannot be created successfully.

Part of Zinc HTTP Components. !

!ZnWebSocketFailed class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 7/19/2012 15:47'!
response: response
	^ self new
		response: response;
		yourself! !

!ZnWebSocketFailed methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 7/19/2012 15:46'!
response
	^ response! !

!ZnWebSocketFailed methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 7/19/2012 15:46'!
response: anObject
	response := anObject! !

Object subclass: #ZnWebSocket
	instanceVariableNames: 'stream role'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-WebSocket-Core'!

!ZnWebSocket commentStamp: '<historical>' prior: 0!
I am ZnWebSocket, a bi-directional, full-duplex communication channel over a single TCP connection.

ZnWebSocket to: 'ws://echo.websocket.org'.
self send: 'Pharo Smalltalk !!' asByteArray.
self read asString.
self close.

===
ZnWebSocket to: 'ws://localhost:1701/ws-echo'.
self sendMasked: 'Pharo Smalltalk !!' asByteArray.
self read asString.
self close.
===

Part of Zinc HTTP Components.!

!ZnWebSocket class methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 12:03'!
isValidWebSocketResponse: client
	| acceptKey response |
	acceptKey := ZnWebSocketUtils handshake: (client request headers at: 'Sec-WebSocket-Key').
	response := client response.
	^ response code = 101
		and: [
			(response headers at: 'Upgrade') asLowercase = 'websocket' 
				and: [
					(ZnWebSocketUtils containsConnectionUpgrade: response headers)
						and: [ 
							(response headers at: 'Sec-WebSocket-Accept') = acceptKey ] ] ]! !

!ZnWebSocket class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 7/19/2012 15:41'!
onStream: socketStream
	^ self new
		onStream: socketStream;
		yourself! !

!ZnWebSocket class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 16:19'!
to: urlObject
	"Attempt to create a new client WebSocket to urlObject.
	Do the initial upgrade handshake and return a functioning ZnWebSocket object.
	Signals a ZnWebSocketFailed error when unsuccessful."
	
	| url client |
	url := urlObject asZnUrl.
	client := self webSocketClientTo: url.
	client execute.
	(self isValidWebSocketResponse: client)
		ifTrue: [
			^ (self onStream: client connection)
				role: #client;
				yourself ]
		ifFalse: [
			client close.
			(ZnWebSocketFailed response: client response) signal ]! !

!ZnWebSocket class methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/22/2012 10:06'!
webSocketClientTo: url
	| client |
	self assert: (#(ws wss) includes: url scheme).
	(client := ZnClient new)
		url: url;
		method: #GET;
		headerAt: 'Upgrade' put: 'websocket';
		headerAt: 'Connection' put: 'Upgrade';
		headerAt: 'Sec-WebSocket-Version' put: '13';
		headerAt: 'Sec-WebSocket-Key' put: ZnWebSocketUtils newClientKey.
	^ client   
	! !

!ZnWebSocket methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 8/21/2012 20:43'!
close
	"Send a close frame and close the connection"

	self sendFrame: ZnWebSocketFrame close.
	stream close.! !

!ZnWebSocket methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 8/21/2012 16:11'!
initialize
	role := #server! !

!ZnWebSocket methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 16:13'!
isClient
	^ role = #client! !

!ZnWebSocket methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 7/19/2012 16:49'!
isConnected
	^ stream notNil and: [ stream isConnected ]! !

!ZnWebSocket methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 16:13'!
isServer
	^ role = #server! !

!ZnWebSocket methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 7/19/2012 15:40'!
onStream: socketStream
	stream := socketStream ! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 20:44'!
ping
	"Send a ping frame."

	self sendFrame: (ZnWebSocketFrame ping: ZnWebSocketUtils newPingPayload)! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 20:44'!
pong: payload
	"Send a pong frame containing payload"
	
	self sendFrame: (ZnWebSocketFrame pong: payload)! !

!ZnWebSocket methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 8/21/2012 16:27'!
printOn: aStream
	super printOn: aStream.
	aStream 
		nextPut: $(; 
		print: role;
		nextPut: $)! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 21:13'!
readFrame
	"Read and return the next non-control frame.
	Control frames are handled. Returns nil when the socket is closed."
	
	| frame |
	frame := ZnWebSocketFrame readFrom: stream.
	frame isControl 
		ifTrue: [
			frame isClose ifTrue: [ 
				self crLog: 'Handling close frame'; close. ^ nil ].
			frame isPing ifTrue: [ 
				self crLog: 'Handling ping frame'; pong: frame body ].
			frame isPong ifTrue: [
				self crLog: 'Handling pong frame' ].
			^ self readFrame ].
	^ frame! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 21:03'!
readMessage
	"Read and return a complete message String or ByteArray, joining frames. Return nil when closed."
	
	| frame bytes isText |
	(frame := self readFrame) ifNil: [ ^ nil ].
	frame isFinal ifTrue: [ ^ frame contents ].
	isText := frame isText.
	bytes := ByteArray new: frame size * 2 streamContents: [ :output |
		frame writeRawBodyOn: output.
		[ 
			(frame := self readFrame) ifNotNil: [ 
				self assert: frame isContinuation. 
				frame writeRawBodyOn: output ].
			frame isNil or: [ frame isFinal ] ] whileFalse ].
	^ isText 
		ifTrue: [ ZnUTF8Encoder new decodeBytes: bytes ]
		ifFalse: [ bytes ]! !

!ZnWebSocket methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 16:11'!
role
	"Return the role of the receiver, either #client or #server"
	
	^ role! !

!ZnWebSocket methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 8/21/2012 16:12'!
role: symbol
	"Set the role of the receiver, either #client or #server"
	
	self assert: (#(client server) includes: symbol).
	role := symbol! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 22:11'!
sendByteFrames: sequenceableCollection
	"Send a binary message consisting of sequenceableCollection byte array frames"

	| frame |
	self assert: sequenceableCollection size > 1.
	(frame := ZnWebSocketFrame bytes: sequenceableCollection first)
		final: false.
	self sendFrame: frame.
	2 to: sequenceableCollection size - 1 do: [ :each |
		(frame := ZnWebSocketFrame bytes: (sequenceableCollection at: each))
			final: false; opcode: 0.
		self sendFrame: frame ].
	(frame := ZnWebSocketFrame bytes: sequenceableCollection last)
		final: true; opcode: 0.
	self sendFrame: frame
	! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 21:18'!
sendBytes: bytes
	"Send bytes in one binary frame"

	| frame |
	frame := ZnWebSocketFrame bytes: bytes.
	self sendFrame: frame! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 16:15'!
sendFrame: frame
	frame masked: self isClient.
	frame writeOn: stream.
	stream flush.! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 21:22'!
sendMessage: stringOrByteArray
	"Send stringOrByteArray in one textual or binary frame"

	stringOrByteArray isString
		ifTrue: [ self sendText: stringOrByteArray ] 
		ifFalse: [ self sendBytes: stringOrByteArray ]! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 20:45'!
sendText: string
	"Send string in one textual frame"

	| frame |
	frame := ZnWebSocketFrame text: string.
	self sendFrame: frame! !

!ZnWebSocket methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 22:11'!
sendTextFrames: sequenceableCollection
	"Send a textual message consisting of sequenceableCollection string frames"

	| frame |
	self assert: sequenceableCollection size > 1.
	(frame := ZnWebSocketFrame text: sequenceableCollection first)
		final: false.
	self sendFrame: frame.
	2 to: sequenceableCollection size - 1 do: [ :each |
		(frame := ZnWebSocketFrame text: (sequenceableCollection at: each))
			final: false; opcode: 0.
		self sendFrame: frame ].
	(frame := ZnWebSocketFrame text: sequenceableCollection last)
		final: true; opcode: 0.
	self sendFrame: frame
	! !

!ZnWebSocket methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 7/20/2012 15:50'!
stream
	^ stream! !

Object subclass: #ZnWebSocketDelegate
	instanceVariableNames: 'prefix handler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-WebSocket-Core'!

!ZnWebSocketDelegate commentStamp: '<historical>' prior: 0!
I am ZnWebSocketDelegate, a plugin to allow a ZnServer to accept incoming WebSocket requests.

When accepted, the request is upgraded and a new WebSocket is created and handed to the handler block.

===
ZnServer startDefaultOn: 1701.
ZnServer default logToTranscript.
ZnServer default delegate
	map: 'ws-test' 
	to: [ :request | ZnResponse ok: (ZnEntity html: ZnWebSocketDelegate wsTestHtml) ];
	map: 'ws-test-local' 
	to: [ :request | ZnResponse ok: (ZnEntity html: ZnWebSocketDelegate wsTestLocalHtml) ];
	map: 'ws-echo'
	to: [ :request | ZnWebSocketDelegate new handleRequest: request ].
ZnServer stopDefault.
===
Part of Zinc HTTP Components.!

!ZnWebSocketDelegate class methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 8/21/2012 11:07'!
wsTestHtml
	^ '<!!DOCTYPE html>
<meta charset="utf-8" />
<title>WebSocket Test</title>

<script language="javascript" type="text/javascript">

  var wsUri = "ws://echo.websocket.org";
  var output;

  function init() {
    output = document.getElementById("output");
    testWebSocket();
  }

  function testWebSocket() {
    websocket = new WebSocket(wsUri);
    websocket.onopen = function(evt) { onOpen(evt) };
    websocket.onclose = function(evt) { onClose(evt) };
    websocket.onmessage = function(evt) { onMessage(evt) };
    websocket.onerror = function(evt) { onError(evt) };
  }

  function onOpen(evt) {
    writeToScreen("CONNECTED");
    doSend("WebSocket rocks");
  }

  function onClose(evt) {
    writeToScreen("DISCONNECTED");
  }

  function onMessage(evt) {
    writeToScreen("<span style=''color: blue;''>RESPONSE: " + evt.data+"</span>");
    websocket.close();
  }

  function onError(evt) {
    writeToScreen("<span style=''color: red''>ERROR:</span> " + evt.data);
  }

  function doSend(message) {
    writeToScreen("SENT: " + message); 
    websocket.send(message);
  }

  function writeToScreen(message) {
    var pre = document.createElement("p");
    pre.style.wordWrap = "break-word";
    pre.innerHTML = message;
    output.appendChild(pre);
  }

  window.addEventListener("load", init, false);

</script>

<h2>WebSocket Test</h2>
<div id="output"></div>

</html> '! !

!ZnWebSocketDelegate class methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 8/21/2012 11:06'!
wsTestLocalHtml
	^ '<!!DOCTYPE html>
<meta charset="utf-8" />
<title>WebSocket Test</title>

<script language="javascript" type="text/javascript">

  var wsUri = "ws://localhost:1701/ws-echo";
  var output;

  function init() {
    output = document.getElementById("output");
    testWebSocket();
  }

  function testWebSocket() {
    websocket = new WebSocket(wsUri);
    websocket.onopen = function(evt) { onOpen(evt) };
    websocket.onclose = function(evt) { onClose(evt) };
    websocket.onmessage = function(evt) { onMessage(evt) };
    websocket.onerror = function(evt) { onError(evt) };
  }

  function onOpen(evt) {
    writeToScreen("CONNECTED");
    doSend("WebSocket rocks");
  }

  function onClose(evt) {
    writeToScreen("DISCONNECTED");
  }

  function onMessage(evt) {
    writeToScreen("<span style=''color: blue;''>RESPONSE: " + evt.data+"</span>");
    websocket.close();
  }

  function onError(evt) {
    writeToScreen("<span style=''color: red''>ERROR:</span> " + evt.data);
  }

  function doSend(message) {
    writeToScreen("SENT: " + message); 
    websocket.send(message);
  }

  function writeToScreen(message) {
    var pre = document.createElement("p");
    pre.style.wordWrap = "break-word";
    pre.innerHTML = message;
    output.appendChild(pre);
  }

  window.addEventListener("load", init, false);

</script>

<h2>WebSocket Test</h2>
<div id="output"></div>

</html> '! !

!ZnWebSocketDelegate methodsFor: 'public' stamp: 'SvenVanCaekenberghe 8/17/2012 14:42'!
handleRequest: request
	"Server delegate entry point"
	
	^ (self isValidWebSocketRequest: request)
		ifTrue: [ self webSocketResponseForRequest: request ]
		ifFalse: [ ZnResponse badRequest: request ]! !

!ZnWebSocketDelegate methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/21/2012 12:03'!
isValidWebSocketRequest: request
	^ request method = #GET
		and: [ 
			(request headers at: 'Upgrade') asLowercase = 'websocket' 
				and: [
					(ZnWebSocketUtils containsConnectionUpgrade: request headers) 
						and: [ 
							(request headers at: 'Sec-WebSocket-Version') = '13'
								and: [ 
									request headers includesKey: 'Sec-WebSocket-Key' ] ] ] ]! !

!ZnWebSocketDelegate methodsFor: 'private' stamp: 'SvenVanCaekenberghe 8/20/2012 17:01'!
responseHeadersForKey: key
	^ ZnHeaders defaultResponseHeaders
			at: 'Upgrade' put: 'websocket';
			at: 'Connection' put: 'Upgrade';
			at: 'Sec-WebSocket-Accept' put: key;
			yourself! !

!ZnWebSocketDelegate methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 8/20/2012 17:02'!
webSocketResponseForRequest: request
	| acceptKey response |
	acceptKey := ZnWebSocketUtils handshake: (request headers at: 'Sec-WebSocket-Key').
	^ ZnWebSocketResponse new
		statusLine: (ZnStatusLine code: 101);
		headers: (self responseHeadersForKey: acceptKey);
		yourself
		! !

Object subclass: #ZnWebSocketFrame
	instanceVariableNames: 'opcode final masked body'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-WebSocket-Core'!

!ZnWebSocketFrame commentStamp: '<historical>' prior: 0!
I am ZnWebSocketFrame, the elementary, low-level unit of data transfer in the WebSocket protocol.

Part of Zinc HTTP Components.!

!ZnWebSocketFrame class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 15:22'!
bytes: byteArray
	^ self new
		opcode: 2;
		final: true;
		body: byteArray;
		yourself! !

!ZnWebSocketFrame class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 15:26'!
close
	^ self new
		opcode: 8;
		final: true;
		yourself! !

!ZnWebSocketFrame class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 20:22'!
ping: bytes
	^ self new
		opcode: 9;
		final: true;
		body: bytes;
		yourself! !

!ZnWebSocketFrame class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 20:22'!
pong: bytes
	^ self new
		opcode: 10;
		final: true;
		body: bytes;
		yourself! !

!ZnWebSocketFrame class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 15:40'!
readFrom: stream
	^ self new
		readFrom: stream;
		yourself! !

!ZnWebSocketFrame class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 8/21/2012 20:30'!
text: string
	^ self new
		opcode: 1;
		final: true;
		text: string;
		yourself! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
body
	^ body! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
body: anObject
	body := anObject! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/22/2012 10:28'!
contents
	self assert: self isControl not.
	^ self isText
		ifTrue: [ self text ]
		ifFalse: [ body ]! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
final
	^ final! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
final: anObject
	final := anObject! !

!ZnWebSocketFrame methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 8/21/2012 15:31'!
initialize
	super initialize.
	final := true! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 16:26'!
isBinary
	^ opcode = 2! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 17:10'!
isClose
	^ opcode = 8! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 17:10'!
isContinuation
	^ opcode = 0! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:46'!
isControl
	^ opcode > 7! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 21:00'!
isEmpty
	^ self size = 0! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 20:34'!
isFinal
	^ final! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 22:29'!
isMasked
	^ masked! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 17:09'!
isPing
	^ opcode = 9! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 17:09'!
isPong
	^ opcode = 10! !

!ZnWebSocketFrame methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/21/2012 16:26'!
isText
	^ opcode = 1! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
masked
	^ masked! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
masked: anObject
	masked := anObject! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
opcode
	^ opcode! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 15:08'!
opcode: anObject
	opcode := anObject! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/22/2012 10:56'!
opcodeName
	| name |
	self assert: (opcode between: 0 and: 15).
	name := #(continuation text binary nil nil nil nil nil close ping pong nil nil nil nil nil) at: opcode + 1.
	^ name ifNil: [ opcode asString ] ifNotNil: [ name ]! !

!ZnWebSocketFrame methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 8/22/2012 10:55'!
printOn: stream
	super printOn: stream.
	stream 
		nextPut: $(; 
		print: self opcodeName;
		space;
		print: self size; nextPut: $B; 
		space;
		nextPutAll: (final ifTrue: [ 'final' ] ifFalse: [ 'non-final' ]);
		nextPut: $)! !

!ZnWebSocketFrame methodsFor: 'private' stamp: 'SvenVanCaekenberghe 8/21/2012 15:32'!
readBodyOfSize: size from: stream
	| mask |
	masked 
		ifTrue: [ 
			mask := stream next: 4.
			body := ByteArray new: size streamContents: [ :out |
				1 to: size do: [ :index |
					out nextPut: (stream next bitXor: (mask at: ((index - 1) \\ 4) + 1)) ] ] ]
		ifFalse: [
			body := stream next: size ]! !

!ZnWebSocketFrame methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 8/22/2012 10:58'!
readFrom: stream
	| byte size mask |
	byte := stream next.
	final := (byte bitAt: 8) = 1.
	opcode := byte bitAnd: 127.
	byte := stream next.
	masked := (byte bitAt: 8) = 1.
	size := byte bitAnd: 127.
	size < 126
		ifFalse: [
			size = 126
				ifTrue: [ size := (stream next: 2) asInteger ]
				ifFalse: [ size := (stream next: 8) asInteger ] ].
	self readBodyOfSize: size from: stream! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 20:59'!
size
	"Return the body size in bytes"
	
	^ body 
		ifNil: [ 0 ] 
		ifNotNil: [ body size ].! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/22/2012 10:28'!
text
	^ ZnUTF8Encoder new decodeBytes: body! !

!ZnWebSocketFrame methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 20:29'!
text: string
	self body: (ZnUTF8Encoder new encodeString: string)
! !

!ZnWebSocketFrame methodsFor: 'writing' stamp: 'SvenVanCaekenberghe 8/21/2012 21:02'!
writeBodyOfSize: size on: stream
	| mask |
	masked
		ifTrue: [
			mask := ZnWebSocketUtils newMask.
			stream nextPutAll: mask ].
	size = 0 ifTrue: [ ^ self ].
	masked 
		ifTrue: [
			body doWithIndex: [ :each :index |
				stream nextPut: (each bitXor: (mask at: ((index - 1) \\ 4) + 1)) ] ] 
		ifFalse: [ self writeRawBodyOn: stream ]! !

!ZnWebSocketFrame methodsFor: 'writing' stamp: 'SvenVanCaekenberghe 8/22/2012 10:58'!
writeOn: stream
	| byte size |
	byte := (final ifTrue: [ 128 ] ifFalse: [ 0 ]) + opcode.
	stream nextPut: byte.
	byte := (masked ifTrue: [ 128 ] ifFalse: [ 0 ]).
	(size := self size) < 126
		ifTrue: [ stream nextPut: (byte + size) ]
		ifFalse: [
			size < 65536
				ifTrue: [ stream nextPut: (byte + 126); nextPutAll: (size asByteArrayOfSize: 2) ]
				ifFalse: [
					self assert: size < (2 raisedTo: 31).
					stream nextPut: (byte + 127); nextPutAll: (size asByteArrayOfSize: 8) ] ].
	self writeBodyOfSize: size on: stream ! !

!ZnWebSocketFrame methodsFor: 'writing' stamp: 'SvenVanCaekenberghe 8/21/2012 21:02'!
writeRawBodyOn: stream
	self isEmpty
		ifFalse: [ stream nextPutAll: body ]! !

Object subclass: #ZnWebSocketUtils
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-WebSocket-Core'!

!ZnWebSocketUtils commentStamp: '<historical>' prior: 0!
I am ZnWebSocketUtils.

Part of Zinc HTTP Components.!

!ZnWebSocketUtils class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 12:02'!
containsConnectionUpgrade: headers
	^ (((headers at: 'Connection') 
			findTokens: ',') 
				collect: [ :each | each trimBoth asLowercase ]) 
					includes: 'upgrade'! !

!ZnWebSocketUtils class methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 7/19/2012 15:07'!
handshake: clientKey
	"Perform the WebSocket key handshake"
	
	"self handshake: 'dGhlIHNhbXBsZSBub25jZQ=='"
	
	| string hash |
	string := clientKey, self serverGUID.
	hash := SHA1 hashMessage: string.
	^ ZnUtils encodeBase64: hash! !

!ZnWebSocketUtils class methodsFor: 'protocol' stamp: 'SvenVanCaekenberghe 7/20/2012 15:06'!
newClientKey
	"Generate a new WebSocket client key"
	
	| rawKey |
	rawKey := self randomByteArrayOfSize: 16.
	^ ZnUtils encodeBase64: rawKey! !

!ZnWebSocketUtils class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 7/20/2012 15:06'!
newMask
	^ self randomByteArrayOfSize: 4! !

!ZnWebSocketUtils class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 20:23'!
newPingPayload
	^ self randomByteArrayOfSize: 4! !

!ZnWebSocketUtils class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 7/20/2012 15:06'!
randomByteArrayOfSize: size
	^ ByteArray new: size streamContents: [ :stream |
		size timesRepeat: [ 
			stream nextPut: 256 atRandom - 1 ] ].
! !

!ZnWebSocketUtils class methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 7/19/2012 14:54'!
serverGUID
	"This is the special WebSocket server GUID"
	
	^ '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'! !

ZnResponse subclass: #ZnWebSocketResponse
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-WebSocket-Core'!

!ZnWebSocketResponse commentStamp: '<historical>' prior: 0!
I am ZnWebSocketResponse used to continue the WebSocket conversation after the connection upgrade.
I am a ZnResponse.

Part of Zinc HTTP Components.!

!ZnWebSocketResponse methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 8/21/2012 21:19'!
useConnection: connection
	"Hook method overwritten to give the receiver the chance to 
	keep using connection in the current thread/process after the server wrote the response.
	We start a WebSocket conversation echoing incoming messages."
	
	| webSocket message |
	webSocket := ZnWebSocket onStream: connection.
	[ 
	[
		[
			self crLog: 'Reading message...'.
			(message := webSocket readMessage) ifNil: [ ^ self crLog: 'Received close, done' ].
			self crLog: 'Received message: ', message, ' echoing...'.
			webSocket sendMessage: message.
			self crLog: 'Message sent' ] on: ConnectionTimedOut do: [ self crLog: 'Ignoring connection timeout, looping' ]
	] repeat ] on: ConnectionClosed do: [ self crLog: 'Ignoring connection closed, done' ]! !

!ZnWebSocketResponse methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/20/2012 16:58'!
wantsConnectionClose
	"After we are done, the server can close the connection we were using"
	 
	^ true! !
